; *****************************************************************************
; *****************************************************************************
;
;       File    : Sound.Asm
;       Author  : Paul Robson (autismuk@aol.com)
;       Purpose : Adlib Soundcard Interface
;       Created : 9th October 1997
;       Modified: 17th October 1997
;
;       Functions :
;               NNewFrame       Start a New frame
;               NToggle         Sound bit toggled.
;               NEndFrame       End a frame
;               DetectAdlib     Detect the SoundCard
;               NSoundOff       Turn the Sound Off
;
; *****************************************************************************
; *****************************************************************************

.386p 
.model flat
include extern.inc
.data
SoundCount dd   0                       ; Number of Sound Toggles this Frame
SCFirst    dd   0                       ; Cycle count (first cycle)
SCLast     dd   0                       ; Cycle count (last cycle)
SCPrev     dd   0                       ; Cycle count (previous cycle)
OldVolume  db   255,0,0,0               ; Sound Channel Volume
ChanToggle db   0,0,0,0                 ; Selected Channel
SoundOut   dd   0
.code

; *****************************************************************************
;
;                               Turn all noise off
;
; *****************************************************************************

NSoundOff Proc
        push    eax
        xor     ax,ax                   ; Channel 0 to 'off'
        call    SetVolume
        inc     al
        call    SetVolume
        mov     [OldVolume],255         ; Reset old volume
        pop     eax
        ret
EndP NSoundOff

; *****************************************************************************
;
;                  Start tracking sound for a new channel
;
; *****************************************************************************
       
NNewFrame Proc
        mov     [SoundCount],0
        mov     [ChanToggle],0
        ret
EndP NNewFrame

; *****************************************************************************
;
;                   Report toggling of the speaker port
;
; *****************************************************************************

NToggle Proc
        push    eax
        mov     eax,[SCLast]            ; save previous
        mov     [SCPrev],eax
        mov     eax,[Cycles]            ; get t-states
        mov     [SCLast],eax            ; Save as 'last' sound
        inc     [SoundCount]            ; Increment sound counter value
        cmp     [SoundCount],1          ; if this is the first save that too
        jne     _NTG1
        mov     [SCFirst],eax           ; save the first t-state for sound
        jmp     _NTG0

_NTG1:  cmp     [SoundMode],smMean      ; if mean pitch sound mode
        je      _NTG0                   ; update channel at the end...
        mov     eax,[SCLast]
        sub     eax,[SCPrev]
        mov     ebx,2
        mov     cl,[ChanToggle]
        cmp     [SoundMode],smChan1     ; if single channel mode
        je      _NTG2                   ; skip
        xor     [ChanToggle],1          ; toggling the channel.
_NTG2:  call    NSetChannel

_NTG0:  mov     [SoundOut],7
        pop     eax
        ret
EndP NToggle

; *****************************************************************************
;
;           Calculate sound for end of frame & make the noise
;
; *****************************************************************************

NEndFrame Proc
        mov     eax,[ScLast]            ; eax = total difference
        sub     eax,[ScFirst]
        mov     ebx,[SoundCount]        ; ebx = number of toggles
        mov     ecx,0                   ; ecx = channel.

        cmp     ebx,4
        jb      _NEF0
        cmp     [SoundMode],smMean      ; mean pitch ?
        jne     _NEF1                   ; if not don't set the channel
        call    NSetChannel             ; pitch
        ret

_NEF0:  dec     [SoundOut]
        jnz     _NEF1
        call    NSoundOff
_NEF1:  ret
EndP NEndFrame

; *****************************************************************************
;
;       Set the frequency of channel CL to EBX toggles over EAX cycles.
;       A toggle is when the EAR bit changes (which is why there is a
;       DEC EBX - a single sine wave as 3 crossing points).
;
; *****************************************************************************

NSetChannel Proc
        or      eax,eax                 ; check difference non zero
        jz      _NSC0
        xor     edx,edx                 ; eDX:eAX = cycle count diff
        dec     ebx                     ; dec toggle count (think about it)
        idiv    ebx                     ; eAX = Cycle Diff / # Toggles

;
;       there have been "SoundCount" changes of state between
;       in the given time. Now EAX contains the average
;       number of Cycles per Toggle.
;
        cmp     eax,32768               ; 32768 cycles. This corresponds
        jae     _NSC0                   ; to 3580000/(32768*2) Hz
                                        ; Less than this is silence

        mov     ebx,eax                 ; frequency from 0-32768 to
        shr     ebx,3                   ; 0-4095
        and     ebx,4095
        mov     al,cl                   ; al = channel number
        call    SetFrequency
        mov     ah,[Volume]             ; set volume to (whatever)
        jmp     _NSCOut

_NSC0:  xor     ax,ax                   ; Silence....

_NSCOut:cmp     ah,[OldVolume]          ; only update volume if changed
        je      _NSC1
        mov     [OldVolume],ah
        mov     al,cl
        call    SetVolume
_NSC1:  ret
EndP NSetChannel

; ***********************************************************************
; ***********************************************************************
; *                                                                     *
; *               Low level interface to Adlib card                     *
; *                Nicked from my Gameboy Emulator                      *
; *                                                                     *
; * DetectAdlibCard()            Check existence of card & set it up    *
; * SetVolume(AL=Channel,AH=Vol) Set Channel AL to volume AH (0..15)    *
; * SetFrequency(AL=Chan,BX=Frq) Set Channel AL to Freq BX (0..2047)    *
; *                                                                     *
; ***********************************************************************
; ***********************************************************************

.data
AdlibFound      db      0,0,0,0                 ; 1 if card found,0 otherwise
FMPort          dw      0388h                   ; Adlib FM Port
.code

; ***********************************************************************
; *                                                                     *
; *                 Write AL to ADLIB Register AH                       *
; *                                                                     *
; ***********************************************************************

WriteFM Proc
        push    eax ebx ecx edx
        mov     dx,[FMPort]             ; dx points to adlib card
        xchg    al,ah
        out     dx,al                   ; output the address
        mov     ecx,256                 ; wait a bit
WFM1:   dec     ecx
        jnz     WFM1
        xchg    al,ah                   ; output the data
        inc     dx
        out     dx,al
        mov     ecx,768                 ; wait a bit again
WFM2:   dec     ecx
        jnz     WFM2
        pop     edx ecx ebx eax
        ret
EndP WriteFM

; ***********************************************************************
; *                                                                     *
; *                 Read Adlib Status Byte to AL                        *
; *                                                                     *
; ***********************************************************************

ReadFM  Proc
        push    edx
        mov     dx,[FMPort]
        in      al,dx
        pop     edx
        ret
EndP ReadFM

; ***********************************************************************
; *                                                                     *
; *                 Detect & Initialise Adlib card                      *
; *                                                                     *
; ***********************************************************************

DetectAdlib Proc
        push    eax ebx ecx edx
        mov     [AdlibFound],0                  ; not found
        mov     ax,0100h                        ; writefm(1,0)
        call    WriteFM                         ; disable test register
        mov     ax,0460h                        ; writefm(4,60)
        call    WriteFM                         ; disable timer/stop
        mov     ax,0480h                        ; writefm(4,80)
        call    WriteFM                         ; delete status register
        call    ReadFM                          ; read the card
        and     al,0E0h                         ; x & 0xE0 must be zero
        jnz     DetectExit
        mov     ax,02FFh                        ; writefm(2,ff)
        call    WriteFM                         ; counter 1 = 255
        mov     ax,0421h                        ; writefm(4,21)
        call    WriteFM                         ; connection on/start timer
        mov     cx,200                          ; wait a bit
DAD1:   call    ReadFM                          ; read the port 200 times
        dec     cx
        jnz     DAD1
        and     al,0E0h                         ; must be 0xC0
        cmp     al,0C0h
        jne     DetectExit
        mov     [AdlibFound],1                  ; we've found it
        mov     ax,0460h                        ; writefm(4,60)
        call    WriteFM                         ; disable timer/stop
        mov     ax,0480h                        ; writefm(4,80)
        call    WriteFM                         ; delete status register
        mov     ax,0800h                        ; writefm(8,0)
        call    WriteFM                         ; melody mode on
        mov     ax,0BD00h                       ; writefm(BD,0)
        call    WriteFM                         ; vibrato off

        mov     ebx,offset AdlibSetupData       ; initialise the card
SIN1:   mov     ah,[ebx+0]                      ; send out a stream of
        mov     al,[ebx+1]                      ; data to the card setting
        call    WriteFM                         ; up the channels
        add     ebx,2
        mov     al,[ebx+0]                      ; terminated by zero
        cmp     al,0
        jnz     SIN1
DetectExit:
        pop     edx ecx ebx eax
        ret
EndP DetectAdlib

; ***********************************************************************
; *                                                                     *
; *               Set Frequency of Channel AL to BX                     *
; *                                                                     *
; ***********************************************************************

SetFrequency Proc
        push    eax ebx
        cmp     [AdlibFound],0                  ; check card
        jz      SFQExit
        cmp     bx,4096                         ; must be 0..2047
        jae     SFQExit
        mov     ah,al                           ; ah = Channel+0A0h
        add     ah,0A0h
        movzx   ebx,bx                          ; extend bx to 32 bits

        mov     al,[byte ptr FreqData+ebx*2+1]
        or      al,al
        jz      SFQExit

        mov     al,[byte ptr FreqData+ebx*2]    ; read the frequency (low)
        call    WriteFM                         ; WriteFM(Channel+0xA0,low)
        add     ah,010h                         ; ah = Channel+0B0h
        mov     al,[byte ptr FreqData+ebx*2+1]  ; read the frequency (high)
        call    WriteFM
SFQExit:pop     ebx eax
        ret
EndP SetFrequency

; ***********************************************************************
; *                                                                     *
; *                  Set Volume of Channel AL to AH                     *
; *                                                                     *
; ***********************************************************************

SetVolume Proc
        push    eax ebx ecx
        movzx   ebx,ah                          ; ebx = 32 bit volume
        cmp     ebx,16                          ; must be in range 0..15
        jae     SVLExit
        mov     cl,[byte ptr VolData+ebx]       ; cl = volume value
        mov     bl,al                           ; ebx = 32 bit channel
        mov     ah,[byte ptr InstAddr+ebx]      ; ah = register to write
        add     ah,043h                         ; = instad[]+0x43
        mov     al,63                           ; al = 63 - volume
        sub     al,cl
        call    WriteFM                         ; write the volume
SVLExit:pop     ecx ebx eax
        ret
EndP SetVolume

.data
AdlibSetupData:                                 ; ADLIB Card setup data
        db      32,35
        db      35,33
        db      64,30
        db      67,0
        db      96,255
        db      99,255
        db      128,8
        db      131,8
        db      224,0
        db      227,0
        db      192,14
        db      33,35
        db      36,33
        db      65,30
        db      68,0
        db      97,255
        db      100,255
        db      129,8
        db      132,8
        db      225,0
        db      228,0
        db      193,14
        db      34,35
        db      37,33
        db      66,30
        db      69,0
        db      98,255
        db      101,255
        db      130,8
        db      133,8
        db      226,0
        db      229,0
        db      194,10
        db      40,46
        db      43,46
        db      72,64
        db      75,3
        db      104,255
        db      107,255
        db      136,15
        db      139,15
        db      232,0
        db      235,3
        db      195,14
        db      0,0
FreqData:                               ; table of frequency values
        include Freq.Inc

InstAddr: db 0,1,2,8                    ; Instrument addresses
VolData:  db    0,35,37,39,41,43,45,47  ; Volume Levels
          db    49,51,53,55,57,59,61,63
end
