/********************************************************************/
/********************************************************************/
/*																	*/
/*							Utilities               				*/
/*																	*/
/*				General purpose utility functions					*/
/*																	*/
/********************************************************************/
/********************************************************************/

#include "z80.h"
#include <stdarg.h>

/********************************************************************/
/*																	*/
/*				Allocate Labels from the Label List					*/
/*																	*/
/********************************************************************/

static int NextFreeLabel = 10000;		/* Next available label */

int	Alloc(int Labels)
{
ASSERT(Labels > 0 && Labels <= 4);
NextFreeLabel = NextFreeLabel+Labels;	/* Allocate them */
return(NextFreeLabel-Labels);			/* Return the number of the first */
}

/********************************************************************/
/*																	*/
/*			Return the text value of a system label					*/
/*																	*/
/********************************************************************/

static char SLBuffer[16];

char *SysLabel(int Label)
{
ASSERT(Label >= 10000);
sprintf(SLBuffer,"__Sys%05d",Label);
return(SLBuffer);
}

/********************************************************************/
/*																	*/
/*				Return label corresponding to an opcode				*/
/*																	*/
/********************************************************************/

static char OPBuffer[24];

char *OpLabel(int Group,int Op)
{
ASSERT(ISGROUP(Group));
ASSERT(Op >= 0 && Op < 256);
sprintf(OPBuffer,"_OG%d_%02x",Group,Op);
return(OPBuffer);
}

/********************************************************************/
/*																	*/
/*						Define a new group							*/
/*																	*/
/********************************************************************/

static BOOL bInDef = FALSE;				/* TRUE when in definition */
static BOOL bDefined[8][256] = {FALSE};	/* TRUE when instructions defined */

void Define(int Group,int Op,int Cycles,char *Format,...)
{
va_list AList;
char Buffer[128];

va_start(AList,1);						/* Build & Display the Comment */
vsprintf(Buffer,Format,AList);
va_end(AList);
Comment(Buffer);

ASSERT(ISGROUP(Group));					/* Validation first */
ASSERT(Op >= 0 && Op < 256);
ASSERT(!bInDef);
ASSERT(!bDefined[Group][Op]);

printf("align dword\n");
printf("%s:\n",OpLabel(Group,Op));		/* Label */
if (Cycles != 0) AddCycles(Cycles);		/* Cycles */
bDefined[Group][Op] = TRUE;				/* Defined instruction */
bInDef = TRUE;							/* In Definition */
}

/********************************************************************/
/*																	*/
/*						  End of Definition							*/
/*																	*/
/********************************************************************/

void EndDefine(void)
{
ASSERT(bInDef);							/* Check opened ! */
bInDef = FALSE;
printf("    ret\n\n");
}

/********************************************************************/
/*																	*/
/*					Generate code to add cycles						*/
/*																	*/
/********************************************************************/

void AddCycles(int Cycles)
{
ASSERT(Cycles > 0 && Cycles <= 40);
printf("    add [dword ptr Cycles],%d\n",Cycles);
}

/********************************************************************/
/*																	*/
/*					Terminate program with error					*/
/*																	*/
/********************************************************************/

void Abort(int Line,char *File)
{
fprintf(stderr,"*** ERROR : %s (%d) ***\n\n",File,Line);
exit(-1);
}

/********************************************************************/
/*																	*/
/*			 		Print a StarBar comment                  		*/
/*																	*/
/********************************************************************/

void Comment(char *Comment)
{
char s1[128],s2[128];
int n;
BOOL bMajor = FALSE;

ASSERT(ISSTRING(Comment));
if (*Comment == '*') Comment++,bMajor = TRUE;
n = 33-strlen(Comment)/2;               /* Centre text in a line */
sprintf(s1,"%*s%s%*s",n,"",Comment,n,"");
s1[65] = '\0';							/* Make it 65 chars long */
s2[0] = '\0';							/* Build a line of * that size */
while (strlen(s2) < strlen(s1)) strcat(s2,"*");
if (bMajor) printf("; ***%s***\n",s2);	/* Display the comment */
printf("; ***%s***\n",s2);
printf("; ***%s***\n",s1);
if (bMajor) printf("; ***%s***\n",s2);
printf("; ***%s***\n",s2);
}

/********************************************************************/
/*																	*/
/*			 Check if string in Colon seperated string				*/
/*																	*/
/********************************************************************/

BOOL InString(char *String,char *List)
{
int  Size;
ASSERT(ISSTRING(String));				/* Checking first */
ASSERT(ISSTRING(List));
Size = strlen(String);					/* Length of string */
while (*List != '\0')					/* Go until the end */
	{
	if (strnicmp(List,String,Size) == 0	/* if start of string is xxxx: ok */
		&& List[Size] == ':') return(TRUE);
	while (*List != ':' &&				/* Find : (or end of string) */
					*List != '\0') List++;
	if (*List == ':') List++;			/* if : found skip over it */
	}
return(FALSE);
}

/********************************************************************/
/*																	*/
/*						Output Jump Tables							*/
/*																	*/
/********************************************************************/

void GenerateJumpTables(void)
{
int i,j,n;
static char *Desc[] = { "Normal","CB","ED","DD/FD","DDCB/FDCB" };

for (i = 0;i < SECTIONS;i++)
	{
	n = 0;
	printf("_JumpTable%d:\n",i);
	for (j = 0;j < 256;j++)
		{
		if (j % 8 == 0)
			printf("    dd ");
		printf("%s",bDefined[i][j] ? OpLabel(i,j) : "BError ");
		if (bDefined[i][j]) n++;
		printf( (j % 8 == 7) ? "\n" : ",");
		}
	fprintf(stderr,"Table \"%s\" generated %d out of 256.\n",Desc[i],n);
	}
}

/********************************************************************/
/*																	*/
/*						Dispatcher Code   							*/
/*																	*/
/********************************************************************/

void Dispatcher(int Table,BOOL bIncR)
{
if (bIncR) printf("    inc [byte ptr _R]\n");
printf("    xor eax,eax\n");
Fetch("al");
printf("    jmp [dword ptr _JumpTable%d+eax*4]\n",Table);
}